/**
 * Module Dependencies
 */

const Custom = require('custom-error-instance')
const log = require('loo')('itty-postgres')
const template = require('colon-template')
const Environment = require('envobj')
const exp = require('./exp')
const pg = require('pg')

/**
 * Postgres Error
 */

const PostgresError = Custom('PostgresError')

/**
 * Environment Variables
 */

const env = Environment({
  POSTGRES_URL: String
})

/**
 * Exports
 */

module.exports = exp({ multi, query, one, raw })

/**
 * Run multiple commands with one client
 */

function * multi (fn) {
  const client = yield connect()
  let ret = yield fn({
    query: queryAll(client),
    one: queryOne(client),
    raw: queryRaw(client)
  })
  yield disconnect(client)
  return ret
}

/**
 * Query
 */

function * query (query, params) {
  const client = yield connect()
  const ret = yield queryAll(client)(query, params)
  yield disconnect(client)
  return ret
}

/**
 * One
 */

function * one (query, params) {
  const client = yield connect()
  const ret = yield queryOne(client)(query, params)
  yield disconnect(client)
  return ret
}

/**
 * Raw
 */

function * raw (query) {
  const client = yield connect()
  const ret = yield queryRaw(client)(query)
  yield disconnect(client)
  return ret
}

/**
 * Query All
 */

function queryAll (client) {
  return function (query, params) {
    return new Promise((resolve, reject) => {
      query = prepare(query, params)
      log('query').info(query, params)
      client.query(query, function (err, data) {
        if (err) return reject(error(err, query))
        resolve(data.rows || [])
      })
    })
  }
}

/**
 * Query One
 */

function queryOne (client) {
  return function (query, params) {
    return new Promise((resolve, reject) => {
      query = prepare(query, params)
      log('one').info(query, params)
      client.query(query, function (err, data) {
        if (err) return reject(error(err, query))
        const rows = data.rows
        resolve(rows && rows.length ? rows[0] : null)
      })
    })
  }
}

/**
 * Query Raw
 */

function queryRaw (client) {
  return function (query) {
    return new Promise((resolve, reject) => {
      log('raw').info(query)
      client.query(query, function (err, data) {
        if (err) return reject(error(err, query))
        resolve(data.rows || [])
      })
    })
  }
}
/**
 * Connect
 */

function connect (url) {
  url = url || env.POSTGRES_URL
  return function (fn) {
    const client = new pg.Client(url)
    client.connect(fn)
    return client
  }
}

/**
 * Disconnect the client
 */

function disconnect (client) {
  return function (fn) {
    client.end()
    client.once('end', fn)
  }
}

/**
 * Create an Error
 *
 * @param {Error} err
 * @param {Object} query
 */

function error (err, query) {
  let code = query.text ? query.text : query
  code = code.trim ? code.trim() : code
  return new PostgresError({
    message: err.message,
    code: code
  })
}

/**
 * Prepare a query
 *
 * @param {String} query
 * @param {Object} params
 * @return {Object}
 */

function prepare (query, params) {
  params = params || {}
  var values = []
  params = Object.keys(params || {}).reduce(function (params, key, i) {
    values.push(params[key])
    params[key] = '$' + (i + 1)
    return params
  }, params)

  return {
    text: template(query, params),
    values: values
  }
}
